#!/usr/bin/python3

"""
Name: gsettings-datetime.py - [The GSettings Project](https://gsettings.tuxfamily.org)
Brief: clock (date and time) setup
Author: Frank ENDRES <frankendres.tuxfamily.org> (disabled <first@last.fr>
Version: v0.5
License: CeCILL - http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html (GPL compatible)
"""

from typing import List
import os
import subprocess
import datetime
import time
import gettext; _ = gettext.gettext
import gi
gi.require_version('Gtk', '3.0')
# ~ gi.require_version('Gtk', '4.0')
from gi.repository import GLib, Gtk, Gio
import warnings; warnings.filterwarnings("ignore", category=DeprecationWarning) #Gtk.EntryCompletion





# GTK3 / 4 compatibility layer -> code is written in GTK4 style
# to toogle with Geany: select the whole Compat bloc, then <Ctrl> + <e>
class CompatApplicationWindow(Gtk.ApplicationWindow): #GTK3
	def set_child(self, child): #renamed method
		for content in self.get_children(): #keep header bar and remove other content
			if not isinstance(content, CompatHeaderBar):
				self.remove(content)
		self.add(child)

	def present(self): #renamed method
		self.show_all()
		super().present()
# ~ class CompatApplicationWindow(Gtk.ApplicationWindow):
	# ~ pass #GTK4


class CompatBox(Gtk.Box): #GTK3
	def append(self, widget): #renamed method
		self.pack_start(widget, False, False, 0)
# ~ class CompatBox(Gtk.Box): #GTK4
	# ~ pass


class CompatRadioButton(Gtk.RadioButton): #GTK3
	def set_group(self, radio): #renamed method
		self.join_group(radio)
# ~ class CompatRadioButton(Gtk.CheckButton): #GTK4
	# ~ pass


class CompatCalendar(Gtk.Calendar):  #GTK3
	def select_day(self, gDate):
		self.select_month(gDate.get_month() - 1, gDate.get_year())
		super().select_day(gDate.get_day_of_month())
# ~ class CompatCalendar(Gtk.Calendar):  #GTK4
	# ~ pass


class CompatButton(Gtk.Button): #GTK3
	def new_from_icon_name(icon_name): #changed method profile
		return Gtk.Button.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON) 
# ~ class CompatButton(Gtk.Button): #GTK4
	# ~ def new_from_icon_name(icon_name):
		# ~ return Gtk.Button.new_from_icon_name(icon_name)

class CompatAboutDialog(Gtk.AboutDialog): #GTK3
	def set_visible(self, visible): #renamed method
		if visible == True: 
			super().run()
			self.destroy()
# ~ class CompatAboutDialog(Gtk.AboutDialog): #GTK4
	# ~ pass


class CompatHeaderBar(Gtk.HeaderBar): #GTK3
	def __init__(self):
		super().__init__()
		self.set_show_close_button(True)

	def set_title_widget(self, label):
		self.props.title = label.get_label()
# ~ class CompatHeaderBar(Gtk.HeaderBar): #GTK4
	# ~ pass





class DateTimeSettings(Gtk.Application):
	def __init__(self):
		super().__init__(application_id='org.gsettings.datetime') #flags=Gio.ApplicationFlags.DEFAULT_FLAGS
		gettext.textdomain('gsettings-datetime')
		self.connect('activate', self.on_activate)

	def on_activate(self, app):
		self.win = CompatApplicationWindow(application=app)
		self.win.set_default_icon_name('gsettings-datetime')
		header_bar = CompatHeaderBar()
		header_bar.set_title_widget(Gtk.Label(label=_('Clock Configuration')))
		about_button = CompatButton.new_from_icon_name('help-about')
		about_button.connect('clicked', self.show_about_dialog)
		header_bar.pack_start(about_button)
		self.win.set_titlebar(header_bar)
		self.win.set_resizable(False)
		grid = Gtk.Grid()
		grid.set_margin_start(10); grid.set_margin_end(10); grid.set_margin_top(10); grid.set_margin_bottom(10)
		grid.set_row_spacing(10)
		self.win.set_child(grid)

		# NTP
		self.ntp_button = Gtk.CheckButton(label=_('NTP sync'))
		self.ntp_button.set_tooltip_text(_('Enable / disable time synchronization.'))
		grid.attach(self.ntp_button, 1, 0, 1, 1)
		self.ntp_button.connect('toggled', self.on_ntp_clicked)

		# Date
		self.date_label = Gtk.Label(label=_('Date:'))
		grid.attach(self.date_label, 0, 1, 1, 1)
		self.calendar = CompatCalendar()
		grid.attach(self.calendar, 1, 1, 1, 1)

		# Time
		self.time_label = Gtk.Label(label=_('Time:'))
		grid.attach(self.time_label, 0, 2, 1, 1)
		self.time_box = CompatBox(orientation = Gtk.Orientation.HORIZONTAL)
		grid.attach(self.time_box, 1, 2, 1, 1)
		self.hour_spin = Gtk.SpinButton(orientation = Gtk.Orientation.VERTICAL, adjustment = Gtk.Adjustment(lower = 0, upper = 23, step_increment = 1), numeric = True, wrap = True)
		self.time_box.append(self.hour_spin)
		self.time_box.append(Gtk.Label(label = '<span weight="bold"> : </span>', use_markup = True))
		self.minutes_spin = Gtk.SpinButton(orientation = Gtk.Orientation.VERTICAL, adjustment = Gtk.Adjustment(lower = 0, upper = 59, step_increment = 1), numeric = True, wrap = True)
		self.time_box.append(self.minutes_spin)

		# Timezone
		timezone_label = Gtk.Label(label=_('Timezone:'))
		grid.attach(timezone_label, 0, 3, 1, 1)
		self.tz_entry = Gtk.Entry()
		self.tz_entry.connect('changed', lambda e: self.error_label.set_text(''))
		grid.attach(self.tz_entry, 1, 3, 1, 1)
		# Timezone completion
		self.timezones_store = Gtk.ListStore(str)
		for tz in load_tz():
			self.timezones_store.append([tz])
		timezone_completion = Gtk.EntryCompletion() #FIXME: deprecated -> Gtk.SearchBar
		timezone_completion.set_model(self.timezones_store)
		timezone_completion.set_text_column(0)
		timezone_completion.set_match_func(lambda completion, key, itr: key.lower() in self.timezones_store[itr][0].lower())
		self.tz_entry.set_completion(timezone_completion)

		# System time (UTC)
		st_label = Gtk.Label(label=_('System clock:'))
		grid.attach(st_label, 0, 4, 1, 1)
		self.utc_button = CompatRadioButton()
		self.utc_button.set_label(_('UTC'))
		self.utc_button.set_tooltip_text(_('Set hardware clock to UTC (Linux default).'))
		self.local_button = CompatRadioButton()
		self.local_button.set_label(_('localtime'))
		self.local_button.set_tooltip_text(_('Set hardware clock to local time (Windows default).'))
		self.local_button.set_group(self.utc_button)
		st_vbox = CompatBox(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
		st_vbox.append(self.utc_button)
		st_vbox.append(self.local_button)
		grid.attach(st_vbox, 1, 4, 1, 1)

		# Message:
		self.error_label = Gtk.Label()
		grid.attach(self.error_label, 0, 5, 2, 1)
		
		# Buttons
		vbox = CompatBox(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
		grid.attach(vbox, 1, 6, 1, 1)
		reload_button = Gtk.Button(label=_('Reload'))
		reload_button.connect('clicked', self.on_reload_clicked)
		vbox.append(reload_button)
		save_button = Gtk.Button(label=_('Save'))
		save_button.connect('clicked', self.on_save_clicked)
		vbox.append(save_button)

		self.win.present()
		self.on_reload_clicked(reload_button)


	def on_reload_clicked(self, button):
		self.ntp_button.set_active(get_ntp_status())
		year, month, day, hour, minutes, sec, wday, yday, isdst = datetime.datetime.now().timetuple()
		self.calendar.select_day(GLib.DateTime.new_local(year, month, day, hour, minutes, sec))
		self.hour_spin.set_value(hour)
		self.minutes_spin.set_value(minutes)
		self.tz_entry.set_text(get_tz())
		self.tz_entry.set_position(-1)
		self.utc_button.set_active(get_hw_hour() == time.gmtime().tm_hour)
		self.on_ntp_clicked(self.ntp_button)


	def show_about_dialog(self, widget):
		about_dialog = CompatAboutDialog(transient_for=self.win, modal=True)
		about_dialog.set_program_name('GSettings-DateTime')
		about_dialog.set_version('v0.5')
		about_dialog.set_authors(['Frank ENDRES <first@last.fr>'])
		about_dialog.set_translator_credits('\n- de: Frank ENDRES\n- fr: Frank ENDRES')
		about_dialog.set_website('https://gsettings.tuxfamily.org')
		about_dialog.set_license('- CeCILL (Ce[a] C[nrs] I[nria] L[ogiciel] L[ibre])\n- GPL (General Public License)')
		about_dialog.set_logo_icon_name('gsettings-datetime')
		about_dialog.set_visible(True)


	def on_ntp_clicked(self, button):
		if self.ntp_button.get_active():
			self.date_label.set_visible(False)
			date = self.calendar.set_visible(False)
			self.time_label.set_visible(False)
			time = self.time_box.set_visible(False)
		else:
			self.date_label.set_visible(True)
			date = self.calendar.set_visible(True)
			self.time_label.set_visible(True)
			time = self.time_box.set_visible(True)


	def on_save_clicked(self, button):
		ntp = self.ntp_button.get_active()
		set_ntp(ntp)
		tz = self.tz_entry.get_text()
		utc = self.utc_button.get_active()
		if self.check_timezone(tz):
			set_tz(tz)
			if not ntp:
				year, month, day = self.calendar.get_date()
				date = datetime.datetime(year, month + 1, day).strftime("%Y-%m-%d")
				time = datetime.datetime(2001, 1, 1, int(self.hour_spin.get_value()), int(self.minutes_spin.get_value())).strftime("%H:%M")
				set_datetime(date, time)
				set_hwclock(date, time, utc)
			else:
				set_hwclock(datetime.datetime.now().strftime('%Y-%m-%d'), datetime.datetime.now().strftime('%H:%M'), utc)
			app.quit()


	def check_timezone(self, tz: str) -> bool:
		found = False
		iter = self.timezones_store.get_iter_first()
		while not found and iter is not None:
			if tz == self.timezones_store.get_value(iter, 0):
				found = True
			iter = self.timezones_store.iter_next(iter)

		if not found:
			self.error_label.set_markup('<span foreground="red" weight="bold">' + _('Invalid timezone.') + '</span>')

		return found





# Minimalistic TimeZone library (no PyTZ)
ZONE_INFO = '/usr/share/zoneinfo'

def get_tz() -> str:
	tz = None
	if os.path.exists('/etc/localtime-copied-from'): #Slackware
		tz = os.path.realpath('/etc/localtime-copied-from')[len(ZONE_INFO) + 1:]
	elif os.path.exists('/etc/localtime'):
		tz = os.path.realpath('/etc/localtime')[len(ZONE_INFO) + 1:]
	return tz

def load_tz() -> List[str]:
	timezones = []
	_load_tz(ZONE_INFO, timezones)
	return timezones

def _load_tz(folder: str, timezones : List[str]):
	for entry in os.listdir(folder):
		path = os.path.join(folder, entry)
		if not os.path.isdir(path):
			timezones.append(path[len(ZONE_INFO) + 1:])
		elif path != '/usr/share/zoneinfo/right' and path != '/usr/share/zoneinfo/posix':
			_load_tz(path, timezones)

def set_tz(tz: str):
	if os.path.exists('/etc/localtime-copied-from'): #Slackware
		subprocess.run(['cp', '-f', ZONE_INFO + '/' + tz, '/etc/localtime'])
		subprocess.run(['ln', '-sf', ZONE_INFO + '/' + tz, '/etc/localtime-copied-from'])
	else:
		subprocess.run(['ln', '-sf', ZONE_INFO + '/' + tz, '/etc/localtime'])





# Minimalistic time management wrapper
def get_hw_hour() -> int:
	f = open('/sys/class/rtc/rtc0/time', 'r')
	hw_hour = int(f.read()[:2])
	f.close()
	return hw_hour

def set_datetime(date: str, time: str):
	subprocess.run(['date', '-s', date + ' ' + time])

def set_hwclock(date: str, time: str, utc: bool):
	if os.path.exists('/etc/slackware-version'): #Slackware
		f = open('/etc/hardwareclock', 'w')
		f.write('UTC' if utc else 'localtime')
		f.close()
	subprocess.run(['hwclock', '--set', '--date', date + ' ' + time, '--utc' if utc else '--localtime'])





# Minimalistic NTP wrapper
def get_ntp_status() -> bool:
	if os.path.exists('/etc/rc.d/rc.ntpd'): #Slackware
		ntp = os.access('/etc/rc.d/rc.ntpd', os.X_OK)
	elif os.path.exists('/usr/lib/systemd/system/systemd-timesyncd.service'): #SystemD
		ntp = subprocess.run(['systemctl', 'status', 'systemd-timesyncd'], check=False, capture_output=True).returncode == 0
	#elif: #TODO: sysvinit, runit, ...
	else:
		return None
	return ntp

def set_ntp(status: bool):
	if (status != get_ntp_status()):
		if os.path.exists('/etc/rc.d/rc.ntpd'): #Slackware
			if status:
				subprocess.run(['chmod', '+x', '/etc/rc.d/rc.ntpd'])
				subprocess.run(['/etc/rc.d/rc.ntpd', 'start'])
			else:
				subprocess.run(['/etc/rc.d/rc.ntpd', 'stop'])
				subprocess.run(['chmod', '-x', '/etc/rc.d/rc.ntpd'])
		elif os.path.exists('/usr/lib/systemd/system/systemd-timesyncd.service'): #SystemD
			subprocess.run(['systemctl', 'start' if status else 'stop', 'systemd-timesyncd'])
			subprocess.run(['systemctl', 'enable' if status else 'disable', 'systemd-timesyncd'])
		#elif: #TODO: sysvinit, runit, ...





if __name__ == '__main__':
	app = DateTimeSettings()
	app.run([])
